@@ -37,25 +37,32 @@
 #include <glibtop/mem.h>
 #include <glibtop/netload.h>
 
+#if defined __FreeBSD__ || __FreeBSD_kernel__
+#include <netinet/tcp_var.h>
+#endif
+
 #include "session.h"
 #include "flamedefault.h"
 #include "flamecorona.h"
 #include "flameblue.h"
 #include "flamegreen.h"
+#include "flameorange.h"
+#include "flamedarkblue.h"
+#include "flamegreen_msb.h"
+#include "flamecorona_msb.h"
 
-#include "master.xpm"
 #include "icon.xpm"
 
 /******************************************/
 /* Defines                                */
 /******************************************/
 
-#define XMAX 56
-#define YMAX 56
+#define XMAX 57
+#define YMAX 57
 #define CMAPSIZE (XMAX * YMAX)
-#define RGBSIZE (XMAX * YMAX * 3)
+#define RGBSIZE (XMAX * YMAX * 4)
 #define NCOLOURS 256
-#define NFLAMES 4
+#define NFLAMES 8
 
 #define FIRE_NONE	0
 #define FIRE_CPU	1
@@ -75,12 +82,8 @@
 /******************************************/
 
 typedef struct {
-	Display *display;	/* X11 display */
+	GdkDisplay *display;	/* X11 display */
 	GdkWindow *win;		/* Main window */
-	GdkWindow *iconwin;	/* Icon window */
-	GdkGC *gc;		/* Drawing GC */
-	GdkPixmap *pixmap;	/* Main pixmap */
-	GdkBitmap *mask;	/* Dockapp mask */
 
 	int x;			/* Window X position */
 	int y;			/* Window Y position */
@@ -100,18 +103,19 @@ typedef struct {
 /* Functions                              */
 /******************************************/
 
-int update_cpu();
-int update_mem();
-int update_net();
-int update_file();
-int change_cpu(int);
-void change_flame(int);
-GdkCursor *setup_cursor();
-void burn_spot(int, int, int);
-inline void draw_fire(unsigned int);
+static int update_cpu();
+static int update_mem();
+static int update_net();
+static int update_file();
+static int change_cpu(int);
+static void change_flame(int);
+static GdkCursor *setup_cursor();
+static void burn_spot(int, int, int);
+static void draw_fire(unsigned int);
 static void make_wmfire_dockapp();
-void read_config(int, char **);
-void do_help(void);
+void convert_colormaps_to_lsb(void);
+static void read_config(int, char **);
+static void do_help(void);
 
 /******************************************/
 /* Globals                                */
@@ -126,6 +130,10 @@ flame_data fire[] = {
 	{"Coronal", flamecorona},
 	{"Blue", flameblue},
 	{"Green", flamegreen},
+    {"Orange", flameorange},
+    {"Dark Blue", flamedarkblue},
+    {"MSB Green", flamegreen_msb},
+    {"MSB Corona", flamecorona_msb},
 };
 
 int monitor = FIRE_CPU;
@@ -134,6 +142,7 @@ int cpu_av = 1;
 int cpu_id = 0;
 int cpu_nice = 1;
 char net_dev[16] = "ppp0";
+char resource_name[64] = "wmfire";
 int net_spd = 0;
 char *file_name = NULL;
 int file_max = 100;
@@ -141,7 +150,6 @@ int file_min = 0;
 int cmap = 0;
 int lock = 0;
 int proximity = 0;
-int broken_wm = 0;
 
 /******************************************/
 /* Main                                   */
@@ -187,7 +195,8 @@ main(int argc, char **argv)
 				switch (event->type) {
 				case GDK_DESTROY:
 				case GDK_DELETE:
-					gdk_cursor_destroy(cursor);
+					/* No equivalent function in GTK3 but does not seem to affect anything */
+					//gdk_cursor_destroy(cursor);
 					exit(0);
 					break;
 				case GDK_BUTTON_PRESS:
@@ -261,13 +270,42 @@ main(int argc, char **argv)
 
 		usleep(REFRESH);
 
-		/* Draw the rgb buffer to screen */
-		if (!broken_wm)
-			gdk_draw_rgb_image(bm.iconwin, bm.gc, 4, 4, XMAX, YMAX, GDK_RGB_DITHER_NONE, bm.rgb, XMAX * 3);
-		else
-			gdk_draw_rgb_image(bm.win, bm.gc, 4, 4, XMAX, YMAX, GDK_RGB_DITHER_NONE, bm.rgb, XMAX * 3);
+		///* Draw the rgb buffer to screen */
+        int stride;
+        int format = CAIRO_FORMAT_RGB24;
+        
+        /* Copy bm.rgb into a new buffer for Cairo to use */
+        unsigned char data[RGBSIZE] = {0};
+        memcpy(data,bm.rgb,sizeof(unsigned char) * RGBSIZE);
+        
+        /* Initialize Cairo drawing */
+        cairo_rectangle_int_t cairo_rec = {3, 3, 57, 57};
+        cairo_region_t * cairo_region = cairo_region_create_rectangle (&cairo_rec);
+        GdkDrawingContext * gdk_drawing_context = gdk_window_begin_draw_frame (bm.win,cairo_region);
+        cairo_t *cr = gdk_drawing_context_get_cairo_context (gdk_drawing_context);
+        
+        //* Create gray rectangle to draw behind flame as border */
+        cairo_set_line_width (cr, 0.01);
+        cairo_set_source_rgb (cr, 0.75, 0.75, 0.75);
+        cairo_rectangle (cr, 1, 1, 0.1, 0.1);
+        cairo_stroke (cr);
+        cairo_paint (cr);
+                        
+        /* Load pixel map into Cairo surface to draw */
+        cairo_surface_t *surface;
+        stride = cairo_format_stride_for_width (format, XMAX);
+        surface = cairo_image_surface_create_for_data(data, format, XMAX, YMAX, stride );
+        cairo_set_source_surface( cr, surface, 2, 2 );
+        cairo_paint( cr );
+        
+        /* Tell Cairo we're done drawing */
+        gdk_window_end_draw_frame(bm.win, gdk_drawing_context);
+        
+        /* Clean up Cairo structures */
+        cairo_region_destroy(cairo_region);
+        //cairo_destroy( cr );
+        cairo_surface_destroy( surface ); 
 	}
-
 	return 0;
 }
 
@@ -275,7 +313,7 @@ main(int argc, char **argv)
 /* Update cpu load statistics             */
 /******************************************/
 
-int
+static int
 update_cpu()
 {
 	glibtop_cpu cpu;
@@ -316,7 +354,7 @@ update_cpu()
 /* Update memory statistics               */
 /******************************************/
 
-int
+static int
 update_mem()
 {
 	glibtop_mem mem;
@@ -336,12 +374,45 @@ update_mem()
 /* Update network statistics              */
 /******************************************/
 
-int
+#if defined __FreeBSD__ || __FreeBSD_kernel__
+#define GETSYSCTL(name, var) getsysctl(name, &(var), sizeof(var))
+
+static void
+getsysctl(const char *name, void *ptr, size_t len)
+{
+	size_t nlen = len;
+
+	if (sysctlbyname(name, ptr, &nlen, NULL, 0) == -1) {
+		fprintf(stderr, "sysctl(%s...) failed: %s\n", name,
+			strerror(errno));
+		exit(1);
+	}
+	if (nlen != len) {
+		fprintf(stderr, "sysctl(%s...) expected %lu, got %lu\n",
+			name, (unsigned long)len, (unsigned long)nlen);
+		exit(1);
+	}
+}
+#endif
+
+static int
 update_net()
 {
+	int percent;
+
+#if defined __FreeBSD__ || __FreeBSD_kernel__
+	struct tcpstat netload;
+	uint64_t total;
+	static uint64_t oldtotal = 0;
+
+ 	GETSYSCTL("net.inet.tcp.stats", netload);
+	total = netload.tcps_sndbyte + netload.tcps_rcvbyte;
+	percent = 100 * (total - oldtotal) / (net_spd / UPDATE_SEC);
+
+	oldtotal = total;
+#else
 	glibtop_netload netload;
 	static guint64 oldtotal = 0;
-	int percent;
 
 	glibtop_get_netload(&netload,net_dev);
 
@@ -349,6 +420,7 @@ update_net()
 	percent = 100 * (netload.bytes_total - oldtotal) / (net_spd / UPDATE_SEC);
 
 	oldtotal = netload.bytes_total;
+#endif
 
 	return percent;
 }
@@ -357,7 +429,7 @@ update_net()
 /* Update file statistics                 */
 /******************************************/
 
-int
+static int
 update_file()
 {
 	char buf[128];
@@ -370,7 +442,8 @@ update_file()
 	/* First number only. Complex parsing should be done in */
 	/* external program and value saved to monitored file.  */
 
-	fgets(buf, sizeof (buf), fp);
+	if (fgets(buf, sizeof (buf), fp) == NULL)
+		fprintf(stderr, "warning: file '%s' is empty\n", file_name);
 	number = atof(buf);
 	fclose(fp);
 
@@ -383,7 +456,7 @@ update_file()
 /* Change CPU monitor                     */
 /******************************************/
 
-int
+static int
 change_cpu(int which)
 {
 	glibtop_cpu cpu;
@@ -413,7 +486,7 @@ change_cpu(int which)
 /* Change flame effect                    */
 /******************************************/
 
-void
+static void
 change_flame(int which)
 {
 	if (which > 0 && which <= NFLAMES)
@@ -430,24 +503,30 @@ change_flame(int which)
 /* Setup invisible cursor                 */
 /******************************************/
 
-GdkCursor *
+static GdkCursor *
 setup_cursor()
 {
-	GdkPixmap *source, *mask;
-	GdkColor col = { 0, 0, 0, 0 };
-	GdkCursor *cursor;
-	unsigned char hide[] = { 0x00 };
-
-	/* No obviously invisible cursor available though */
-	/* X/GDK, so using a custom 1x1 bitmap instead    */
-
-	source = gdk_bitmap_create_from_data(NULL, hide, 1, 1);
-	mask = gdk_bitmap_create_from_data(NULL, hide, 1, 1);
-
-	cursor = gdk_cursor_new_from_pixmap(source, mask, &col, &col, 0, 0);
-
-	gdk_pixmap_unref(source);
-	gdk_pixmap_unref(mask);
+	GdkDisplay* display = gdk_display_get_default();
+    GdkCursor *cursor;
+    cairo_surface_t *s;
+    cairo_t *cr;
+    GdkPixbuf *pixbuf;
+    
+    s = cairo_image_surface_create (CAIRO_FORMAT_A1, 1, 1);
+    cr = cairo_create (s);
+    cairo_arc (cr, 1.5, 1.5, 1.5, 0, 2 * M_PI);
+    cairo_fill (cr);
+    cairo_destroy (cr);
+    
+    pixbuf = gdk_pixbuf_get_from_surface (s,
+                                          0, 0,
+                                          1, 1);
+    
+    cairo_surface_destroy (s);
+    
+    cursor = gdk_cursor_new_from_pixbuf (display, pixbuf, 0, 0);
+    
+    g_object_unref (pixbuf);
 
 	return cursor;
 }
@@ -456,7 +535,8 @@ setup_cursor()
 /* Burn spot                              */
 /******************************************/
 
-void burn_spot(int x, int y, int c)
+static void
+burn_spot(int x, int y, int c)
 {
 	int i;
 
@@ -470,7 +550,7 @@ void burn_spot(int x, int y, int c)
 /* Draw fire                              */
 /******************************************/
 
-inline void
+static void
 draw_fire(unsigned int load)
 {
 	int x, y, i, j;
@@ -484,7 +564,10 @@ draw_fire(unsigned int load)
 
 	/* Mouse in window */
 	if (proximity) {
-		gdk_window_get_pointer(bm.win, &x, &y, NULL);
+		GdkDisplay* display = gdk_display_get_default();
+        GdkSeat* seat = gdk_display_get_default_seat(display);
+        GdkDevice* pointer = gdk_seat_get_pointer(seat);
+        gdk_window_get_device_position (bm.win, pointer, &x, &y, NULL);
 
 		/* Burn spot at mouse position */
 		if ((y > 1 && y < 53) && (x > 1 && x < 53))
@@ -542,8 +625,10 @@ draw_fire(unsigned int load)
 	}
 
 	/* Convert colourmap to rgb */
+    /* Use i * 4 to index &bm.rgb instead of i * 3 so that the upper 8 bits of the 32 bits
+     * passed to cairo_image_surface_create_for_data() are unused as required by CAIRO_FORMAT_RGB24*/
 	for (i = 0; i < (CMAPSIZE - XMAX); i++)
-		memcpy(&bm.rgb[i * 3], &bm.flame[bm.cmap[i] * 3], 3);
+		memcpy(&bm.rgb[i * 4], &bm.flame[bm.cmap[i] * 3], 3);
 }
 
 /******************************************/
@@ -556,11 +641,9 @@ make_wmfire_dockapp(void)
 #define MASK GDK_BUTTON_PRESS_MASK | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK | GDK_POINTER_MOTION_HINT_MASK
 
 	GdkWindowAttr attr;
-	GdkWindowAttr attri;
 	Window win;
-	Window iconwin;
 
-	GdkPixmap *icon;
+	GdkPixbuf *icon;
 
 	XSizeHints sizehints;
 	XWMHints wmhints;
@@ -569,54 +652,38 @@ make_wmfire_dockapp(void)
 
 	attr.width = 64;
 	attr.height = 64;
-	attr.title = "wmfire";
+	attr.title = resource_name;
 	attr.event_mask = MASK;
 	attr.wclass = GDK_INPUT_OUTPUT;
-	attr.visual = gdk_visual_get_system();
-	attr.colormap = gdk_colormap_get_system();
-	attr.wmclass_name = "wmfire";
+	attr.visual = gdk_screen_get_system_visual (gdk_screen_get_default ());
+	//attr.colormap = gdk_colormap_get_system(); /* Not needed in GTK3? */
+	attr.wmclass_name = resource_name;
 	attr.wmclass_class = "wmfire";
 	attr.window_type = GDK_WINDOW_TOPLEVEL;
 
-	/* Make a copy for the iconwin - parameters are the same */
-	memcpy(&attri, &attr, sizeof (GdkWindowAttr));
-	attri.window_type = GDK_WINDOW_CHILD;
-
 	sizehints.flags = USSize;
 	sizehints.width = 64;
 	sizehints.height = 64;
 
-	bm.win = gdk_window_new(NULL, &attr, GDK_WA_TITLE | GDK_WA_WMCLASS | GDK_WA_VISUAL | GDK_WA_COLORMAP);
+	bm.win = gdk_window_new(NULL, &attr, GDK_WA_TITLE | GDK_WA_WMCLASS | GDK_WA_VISUAL);
 	if (!bm.win) {
 		fprintf(stderr, "FATAL: Cannot make toplevel window\n");
 		exit(1);
 	}
 
-	bm.iconwin = gdk_window_new(bm.win, &attri, GDK_WA_TITLE | GDK_WA_WMCLASS);
-	if (!bm.iconwin) {
-		fprintf(stderr, "FATAL: Cannot make icon window\n");
-		exit(1);
-	}
-
-	win = GDK_WINDOW_XWINDOW(bm.win);
-	iconwin = GDK_WINDOW_XWINDOW(bm.iconwin);
+	win = GDK_WINDOW_XID(bm.win);
 	XSetWMNormalHints(GDK_WINDOW_XDISPLAY(bm.win), win, &sizehints);
 
 	wmhints.initial_state = WithdrawnState;
-	wmhints.icon_window = iconwin;
+	wmhints.icon_window = win;
 	wmhints.icon_x = 0;
 	wmhints.icon_y = 0;
 	wmhints.window_group = win;
 	wmhints.flags = StateHint | IconWindowHint | IconPositionHint | WindowGroupHint;
 
-	bm.gc = gdk_gc_new(bm.win);
-
-	bm.pixmap = gdk_pixmap_create_from_xpm_d(bm.win, &(bm.mask), NULL, master_xpm);
-	gdk_window_shape_combine_mask(bm.win, bm.mask, 0, 0);
-	gdk_window_shape_combine_mask(bm.iconwin, bm.mask, 0, 0);
-
-	gdk_window_set_back_pixmap(bm.win, bm.pixmap, False);
-	gdk_window_set_back_pixmap(bm.iconwin, bm.pixmap, False);
+    /* No equivalent functions seem to exist in GTK3 */
+	//gdk_window_shape_combine_mask(bm.win, bm.mask, 0, 0);
+	//gdk_window_shape_combine_mask(bm.iconwin, bm.mask, 0, 0);
 
 #if 0
         gdk_window_set_type_hint(bm.win, GDK_WINDOW_TYPE_HINT_DOCK);
@@ -625,9 +692,11 @@ make_wmfire_dockapp(void)
         gdk_window_set_skip_taskbar_hint(bm.win, 1);
 #endif
 
-	icon = gdk_pixmap_create_from_xpm_d(bm.win, NULL, NULL, icon_xpm);
-	gdk_window_set_icon(bm.win, bm.iconwin, icon, NULL);
+    icon = gdk_pixbuf_new_from_xpm_data (icon_xpm);
 
+	/* No way to set window icon in GTK3 */
+	//gdk_window_set_icon(bm.win, bm.iconwin, icon, NULL);
+    
 	gdk_window_show(bm.win);
 
 	/* Moved after gdk_window_show due to change in GTK 2.4 */
@@ -638,22 +707,28 @@ make_wmfire_dockapp(void)
 	if (bm.sticky)
 		gdk_window_stick(bm.win);
 #undef MASK
+
 }
 
 /******************************************/
 /* Read config file                       */
 /******************************************/
 
-void
+static void
 read_config(int argc, char **argv)
 {
+    
+    /* Get screen geometry for 'g' */
+    GdkRectangle workarea = {0};
+    gdk_monitor_get_workarea(gdk_display_get_primary_monitor(gdk_display_get_default()), &workarea);
+    
 	int i, j;
 
 	/* Initialise flame to default */
 	bm.flame = fire[cmap].data;
 
 	/* Parse command options */
-	while ((i = getopt(argc, argv, "c:mni:s:xF:H:L:pf:lbhg:yS:")) != -1) {
+	while ((i = getopt(argc, argv, "c:mni:r:s:xF:H:L:pf:lbhg:yS:")) != -1) {
 		switch (i) {
 		case 'S':
 			if (optarg)
@@ -664,9 +739,9 @@ read_config(int argc, char **argv)
 				j = XParseGeometry(optarg, &bm.x, &bm.y, &j, &j);
 
 				if (j & XNegative)
-					bm.x = gdk_screen_width() - 64 + bm.x;
+					bm.x = workarea.width - 64 + bm.x;
 				if (j & YNegative)
-					bm.y = gdk_screen_height() - 64 + bm.y;
+					bm.y = workarea.height - 64 + bm.y;
 			}
 			break;
 		case 'y':
@@ -687,6 +762,10 @@ read_config(int argc, char **argv)
 			if (optarg)
 				strncpy(net_dev, optarg, sizeof (net_dev));
 			break;
+		case 'r':
+			if(optarg)
+				strcpy(resource_name, optarg);
+			break;
 		case 's':
 			if (optarg)
 				net_spd = atoi(optarg);
@@ -721,9 +800,6 @@ read_config(int argc, char **argv)
 		case 'l':
 			lock = 1;
 			break;
-		case 'b':
-			broken_wm = 1;
-			break;
 		case 'h':
 		default:
 			do_help();
@@ -743,14 +819,14 @@ read_config(int argc, char **argv)
 /* Help                                   */
 /******************************************/
 
-void
+static void
 do_help(void)
 {
 	int i;
 
 	fprintf(stderr, "\nWmfire - Flaming Monitor Dock V %s\n\n", VERSION);
 	fprintf(stderr, "Usage: wmfire [ options... ]\n\n");
-	fprintf(stderr, "\t-g [{+-}X{+-}Y]\t\tinital window position\n");
+	fprintf(stderr, "\t-g [{+-}X{+-}Y]\t\tinitial window position\n");
 	fprintf(stderr, "\t-y\t\t\tset window sticky\n");
 	fprintf(stderr, "\t-c [0..%d]\t\tmonitor smp cpu X\n", GLIBTOP_NCPU-1);
 	fprintf(stderr, "\t-m\t\t\tmonitor memory load\n");
@@ -760,12 +836,12 @@ do_help(void)
 	fprintf(stderr, "\t-s [...]\t\tchange network speed (ppp:%dK) (eth:%dM)\n", NET_SPD_PPP, NET_SPD_ETH);
 	fprintf(stderr, "\t-x\t\t\texclude nice'd cpu load\n");
 	fprintf(stderr, "\t-H [...]\t\tset file maximum (high) value\n");
-	fprintf(stderr, "\t-L [...]\t\tset file minumum (low) value\n");
+	fprintf(stderr, "\t-L [...]\t\tset file minimum (low) value\n");
 	fprintf(stderr, "\t-p\t\t\tfire effect only\n");
 	fprintf(stderr, "\t-f [1..%d]\t\tchange flame colour\n\t\t\t\t", NFLAMES);
 	for (i = 0; i < NFLAMES; i++)
 		fprintf(stderr, "%d:%s ", i + 1, fire[i].text);
 	fprintf(stderr, "\n\t-l\t\t\tlock flame colour and monitor\n");
-	fprintf(stderr, "\t-b\t\t\tactivate broken window manager fix\n");
+	fprintf(stderr, "\t-r [...]\t\tchange resource name (default:%s)\n", resource_name);
 	fprintf(stderr, "\t-h\t\t\tprints this help\n");
 }
